! ****************************************************************************
! FILE: mpi_prime.f
! DESCRIPTION:
!   Generates prime numbers.  All tasks distribute the work evenly, taking
!   every nth number, where n is the stride computed as:  (rank *2) + 1
!   so that even numbers are automatically skipped.  The method of using
!   stride is preferred over contiguous blocks of numbers, since numbers
!   in the higher range require more work to compute and may result in
!   load imbalance.  This program demonstrates embarrassing parallelism.
!   Collective communications calls are used to reduce the only two data
!   elements requiring communications: the number of primes found and
!   the largest prime.
! AUTHOR: Blaise Barney 11/25/95 - adapted from version contributed by 
!   Richard Ng & Wong Sze Cheong during MHPCC Singapore Workshop (8/22/95).
! LAST REVISED: 04/02/05
! Modified by A. Emerson 4-12-2014
! ****************************************************************************
! Explanation of constants and variables
!   LIMIT                  = Increase this to find more primes 
!   FIRST                  = Rank of first task 
!   ntasks                 = total number of tasks in partitiion 
!   rank                   = task identifier 
!   n                      = loop variable 
!   pc                     = prime counter
!   pcsum                  = number of primes found by all tasks 
!   foundone               = most recent prime found 
!   maxprime               = largest prime found 
!   mystart                = where to start calculating 
!   stride                 = calculate every nth number

      program prime
      include 'mpif.h'

      integer LIMIT, FIRST
      parameter(LIMIT=2500000)        
      parameter(FIRST=0)

      integer ntasks, rank, ierr, n, pc, pcsum, foundone, maxprime, mystart, stride
      double precision start_time, end_time
      logical result

      call MPI_INIT(ierr)
      call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr)
      call MPI_COMM_SIZE(MPI_COMM_WORLD, ntasks, ierr)
      if ((mod(ntasks,2).ne.0) .or. (mod(LIMIT,ntasks).ne.0)) then
         print *, mod(ntasks,2), mod(LIMIT,ntasks)
         print *,'Sorry -this exercise requires an even number of '
         print *,'processors evenly divisible into ',LIMIT,'.' 
         print *,'Try 4 or 8.'
         call MPI_FINALIZE(ierr)
         stop
      endif

!     Initializations: mystart must be odd number. stride is multiplied
!     by 2 to skip over even numbers.  
      start_time = MPI_WTIME()
      mystart = (rank*2) + 1
      stride = ntasks*2
      pc = 0
      foundone = 0

! ------------------- task with rank 0 does this part ------------------
      if (rank .eq. FIRST) then
         print *,'Using',ntasks,'tasks to scan',LIMIT,'numbers'
!        Assume first four primes are counted here
         pc = 4
!        
!  insert here a do-loop which loops upto ntasks, starting from mystart
!   and stride ntasks   
!   check each n for prime and update counter
!  

!  insert here two Reduces, one to find the sum of the primes, the sencond to find
!  the max prime
      ! call mpi_reduce(pc, pcsum,..)
      ! call mpi_reduce(prime, maxprime,..)

      end_time=MPI_WTIME()
      print *,'Done. Largest prime is ',maxprime,' Total primes ',pcsum
      print *,'Wallclock time elapsed: ' ,end_time-start_time
      endif

! ------------------- all other tasks do this part ---------------------
      if (rank .gt. FIRST) then

!  insert here a do-loop which loops upto ntasks, starting from mystart
!   and stride ntasks   
!   check each n for prime and update counter
!  

!  insert here two Reduces, one to find the sum of the primes, the sencond to find
!  the max prime
      ! call mpi_reduce(pc, pcsum,..)
      ! call mpi_reduce(prime, maxprime,..)


      endif


      call MPI_FINALIZE(ierr)
      end


      subroutine isprime (n,result)
      integer n
      logical result
      integer i, squareroot
      real*4 realn
      if (n .gt. 10) then
         realn = real(n)
         squareroot = int (sqrt(realn))
         do i=3,squareroot,2
           if (mod(n,i) .eq. 0) then
              result = .false.
              return
           endif
         enddo
         result = .true.
         return 
! Assume first four primes are counted elsewhere. Forget everything else
      else 
        result = .false.
        return
      endif
      end
